/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained from:
 *  http://savi.sourceforge.net/
 *  http://www.geom.uiuc.edu/locate/SaVi
 *
 *****************************************************
 *
 * coverage_vis.c
 *
 * Coverage visualization routines.
 *
 * $Id: coverage_vis.c,v 1.19 2004/09/30 20:14:14 lloydwood Exp $
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tk.h>

#include "constants.h"
#include "globals.h"
#include "orbit_utils.h"
#include "savi.h"
#include "stats_utils.h"

#include "coverage_vis.h"

extern Tcl_Interp *interp;	/* Interpreter for this application. */

extern int tcl_script(char[]);

extern int no_access_flag;

static int current_proj = -1; /* whatever the default, we start by drawing */

static char imagename[] = "im_coverage";

static const char format1[] =
  "set coverage \"Current global coverage: %.1f%%\"";
static const char format2[] =
 "set mean_coverage \"Averaged global coverage: %.1f%%\"";

static void write_image_as_ppm(FILE *f);
static void write_image_to_scratchfile(void);

static void set_coverage(int projection, grid *g);

static void coverage_color_copy(void);
static void grid_and_foreground_to_image(int projection, grid * g,
					 int fgOnly);
static int new_foreground(int projection, grid * g, unsigned int force);
static void draw_sin_map_edges(grid * g);
static Tk_PhotoImageBlock *get_image(void);
static void overlay_bitmap(unsigned int h, unsigned int w, unsigned char *b,
			   const char *filename);
static void latlon_to_grid_index(int grid_index[2], LatLon * pl, int projection,
				 grid * g);

static Tk_PhotoImageBlock *image = NULL;
static unsigned char *foreground = NULL;

#define PIXELSIZE 3		/* 3 for RGB */

typedef struct RGB {
    unsigned char red, green, blue;
} RGB;

/* Colors for coverage intensity (in RGB values)  */
/* Any number of colors may be inserted.          */
/* green instead of orange is {255,255,255,
                               0,255,0,
                               0,191,0,
                               0,127,0,
                               0,88,0};
*/

static unsigned char colors[] = { 255, 255, 255, /* no coverage */
				  255, 255, 0,  /* single coverage */
				  255, 127, 0, /* double coverage */
				  255, 47, 0, /* triple coverage */
				  255, 0, 0  /* four or more */
};
#define num_colors sizeof(colors)/sizeof(unsigned char)/3

/* Color for earth outline and ground track */
static unsigned char outline_colors[] = { 180, 180, 180,  /* background */
					  0, 0, 0,	 /* earth outline */
					  0, 255, 0,    /* projected ground track */
					  255, 0, 255, /* selected ground track */
					  0, 255, 0,     /* satellite position */
					  0, 255, 0,    /* rest of cross */
					  255, 0, 255,   /* selected satellite */
					  255, 0, 255   /* rest of cross */
};
#define num_outline_colors sizeof(outline_colors)/sizeof(unsigned char)/3

/* Color for noaccess */
static unsigned char noaccess_colors[] = { 255, 255, 255, /* no intervals missed */
					   0, 0, 255,	 /* one interval missed */
					   0, 0, 217,	/* two intervals missed */
					   0, 0, 180,  /* three intervals missed */
					   0, 0, 143  /* >three intervals missed */
};
#define num_noaccess_colors sizeof(noaccess_colors)/sizeof(unsigned char)/3

static unsigned char * COLDIV = &colors[0];
static unsigned char * COLDEC = &noaccess_colors[0];

/*
 * color_hex - turn combined RRGGBB byte into three RGB bytes.
 */
static void
color_hex(int combined, RGB * color_set) {
  unsigned int rrggbb = (unsigned int) combined;

  color_set->blue = rrggbb & 0x0000FF;
  rrggbb = rrggbb >> 8;
  color_set->green = rrggbb & 0x0000FF;
  rrggbb = rrggbb >> 8;
  color_set->red = rrggbb & 0x0000FF;
}

/*
 * coverage_color_copy - slave color arrays to Tcl values
 */
static void
coverage_color_copy(void)
{
  color_hex(DIV1, (RGB *) &COLDIV[3]);
  color_hex(DIV2, (RGB *) &COLDIV[6]);
  color_hex(DIV3, (RGB *) &COLDIV[9]);
  color_hex(DIV4, (RGB *) &COLDIV[12]);

  color_hex(DEC1, (RGB *) &COLDEC[3]);
  color_hex(DEC2, (RGB *) &COLDEC[6]);
  color_hex(DEC3, (RGB *) &COLDEC[9]);
  color_hex(DEC4, (RGB *) &COLDEC[12]);
}
/*
 * image_init - Set up the Tk image structure based on the coverage grid.
 */
void
image_init(grid * g)
{
  unsigned int height, width, image_size;

  height = g->height;
  width = g->width;
  image_size = height * width;

  image = (Tk_PhotoImageBlock *) malloc(sizeof(Tk_PhotoImageBlock));
  image->pixelPtr = (unsigned char *) malloc(image_size * PIXELSIZE *
					     sizeof(unsigned char *));
  image->width = width;
  image->height = height;
  image->pitch = width * PIXELSIZE;
  image->pixelSize = PIXELSIZE;
  image->offset[0] = 0;
  image->offset[1] = 1;
  image->offset[2] = 2;

  if (!foreground) {
    foreground = (unsigned char *) malloc(image_size * sizeof(unsigned char *));

    cyl_foreground = (unsigned char *) calloc(sizeof(unsigned char *), image_size);
    sin_foreground = (unsigned char *) calloc(sizeof(unsigned char *), image_size);
    unp_foreground = (unsigned char *) calloc(sizeof(unsigned char *), image_size);

    if (width == IMAGE_LARGE_WIDTH) {
      overlay_bitmap(height, width, cyl_foreground, CYLINDRICAL_LARGE_BITMAP_NAME);
      overlay_bitmap(height, width, sin_foreground, SINUSOIDAL_LARGE_BITMAP_NAME);
      overlay_bitmap(height, width, unp_foreground, UNPROJECTED_LARGE_BITMAP_NAME);
    } else if (width == IMAGE_WIDTH) {
      overlay_bitmap(height, width, cyl_foreground, CYLINDRICAL_BITMAP_NAME);
      overlay_bitmap(height, width, sin_foreground, SINUSOIDAL_BITMAP_NAME);
      overlay_bitmap(height, width, unp_foreground, UNPROJECTED_BITMAP_NAME);
    } else {
      error("no appropriately-sized background map for coverage window.");
    }
    draw_sin_map_edges(g);
  }
}

static void
write_image_as_ppm(FILE *f)
{
  unsigned int w, h, i;
  unsigned char c;
  unsigned char *p = image->pixelPtr;

  /* this was grid g-> width, but for temporary convenience... */
  w = Image_Width;
  h = Image_Height;

  /*
   * Write header file describing grid
   */
  fprintf(f,"P6\n");
  fprintf(f,"%i %i\n", w, h);
  fprintf(f,"255\n");

  for (i = 0; i < w * h * PIXELSIZE; i++) {
    c = *p++;
    fputc(c,f);
  }

  fclose(f);
}

static void
write_image_to_scratchfile()
{
  FILE *f;

  if (NULL == (f = fopen("savi-scratch-texturemap.ppm", "w"))) {
    error("unable to open texturemap scratchfile.");
    return;
  }

  write_image_as_ppm(f);
}

void
write_image_to_file(char * name)
{
  FILE *f;

  if (NULL == (f = fopen(name, "w"))) {
    error("unable to open file to save coverage map.");
    return;
  }

  write_image_as_ppm(f);
}

void
reset_foreground(int projection, grid * g)
{
  if (foreground)
    new_foreground(projection, g, TRUE);
}


/*
 * get_image - Return pointer to the image structure.
 */
static Tk_PhotoImageBlock *
get_image(void)
{
  if (!image) {
    error("get_image: image not yet initialized.");
  }

  return image;
}

/*
 * update_display - Update the Tk coverage image, given a completed coverage grid.
 */
void
update_display(int projection, grid * g, int coverage_flag, int tracks_flag,
	       int project_tracks_flag)
{
  Tk_PhotoHandle handle;

  /*
   * Re-write the entire image, unless we are just displaying ground
   * tracks, and there's not a new projection, in which case we can
   * simply write the foreground (which includes ground tracks).
   */
  int fgOnly = !new_foreground(projection, g, FALSE) && !coverage_flag &&
	       !tracks_flag && !project_tracks_flag;
  grid_and_foreground_to_image(projection, g, fgOnly);

  handle = Tk_FindPhoto(interp, imagename);
  if (handle) {
    /* double width of image to tile horizontally */
#if ( (TK_MAJOR_VERSION > 7) && (TK_MINOR_VERSION > 3) )
   /* changed for version 8.4, September 2002. */
    Tk_PhotoPutBlock(handle, get_image(), 0, 0, image->width*2, image->height,
        TK_PHOTO_COMPOSITE_OVERLAY);
#else
    Tk_PhotoPutBlock(handle, get_image(), 0, 0, image->width*2, image->height);
#endif
  } else {
    error("coverage_display: bad image name.");
  }

  /* display coverage percentages */
  set_coverage(projection, g);
}

/*
 * Clear the intensity array.
 */
void
clear_intensity(grid * g)
{
  memset(g->data, 0, (size_t) sizeof(int) * g->height * g->width);
  memset(g->noaccess, 0, (size_t) sizeof(int) * g->height * g->width);
}

/*
 * set_coverage - Display total percent coverage.
 */
static void
set_coverage(int projection, grid * g)
{
  char cmd[sizeof(format1) + 5];

  sprintf(cmd, format1, total_coverage(projection, TRUE, g));
  tcl_script(cmd);
  sprintf(cmd, format2, mean_coverage(projection, g));
  tcl_script(cmd);
}

/*
 * overlay_bitmap - Takes a bitmap and lies it on top of our image array.
 * h,w = height, width of the image array.
 */
static void
overlay_bitmap(unsigned int h, unsigned int w, unsigned char *b,
	       const char filename[])
{
  FILE *f;
  unsigned int pbmtype, image_size, width, height, i, j, c, mask;

  if (NULL == (f = fopen(filename, "r")))
    return;

  forward_over_comments(f);
  fscanf(f, "P%d", &pbmtype);
  forward_over_comments(f);
  fscanf(f, "%d", &width);
  forward_over_comments(f);
  fscanf(f, "%d", &height);
  /* skip one char */
  c = getc(f);

  if ((pbmtype == 4) && (width == w) && (height == h)) {
    image_size = h*w;
    for (i = 0; i < image_size; i += 8) {
      c = getc(f);
      mask = 0x80;
      for (j = 0; j < 8; j++) {
	if (c & mask) {
	  *(b++) = 1;
	} else {
	  b++;
	}
	mask = mask / 2;
      }
    }
  } else {
    fprintf(stderr, "Error in bitmap format: P%d %d %d\n",
		pbmtype, width, height);
  }

  fclose(f);
}

static void
draw_sin_map_edges(grid * g)
{
  SphericalCoordinates point;
  int left[2], right[2];
  unsigned int h = g->height;
  unsigned int w = g->width;
  unsigned int i;

  point.r = 1;
  point.theta = 0;
  point.phi = 0;

  for (i = 0; i < h; i++) {
    intensity_edges(left, right, &point, SINUSOIDAL, g);
    sin_foreground[(w * i + left[0])] = 1;
    sin_foreground[(w * i + right[0])] = 1;
    point.phi += PI / (h - 1);	/* Increment phi by one pixel */
  }
}

/*
 * new_foreground - Initialize foreground intensity image, including earth outline.
 * Returns TRUE if foreground must be re-drawn, FALSE otherwise.
 */
static int
new_foreground(int projection, grid * g, unsigned int force)
{
  unsigned int image_size;

  if (!force && (projection == current_proj))
    return (FALSE);

  if (current_proj != projection) {
    current_proj = projection;
    clear_intensity(g);
  }

  image_size = g->height * g->width;

  if (projection == CYLINDRICAL) {
    Longitude_Center_Line = LONGITUDE_CENTER_LINE;
    memcpy(foreground, cyl_foreground, image_size);
  } else if (projection == UNPROJECTED) {
    Longitude_Center_Line = LONGITUDE_CENTER_LINE;
    memcpy(foreground, unp_foreground, image_size);
  } else if (projection == SINUSOIDAL) {
    Longitude_Center_Line = LONGITUDE_CENTER_LINE_SINU;
    memcpy(foreground, sin_foreground, image_size);
  }
  return (TRUE);
}

/*
 * Copy foreground (land masses and ground tracks, if any) only onto image.
 * The routine re-draws an image pixel only for corresponding foreground array
 * elements which are non-zero.
 */
static void
grid_and_foreground_to_image(int projection, grid * g, int fgOnly)
{
  unsigned int i, j, offset, surround, ci;
  unsigned int w = g->width;
  unsigned int h = g->height;
  int drawWidth = w;
  unsigned char *c;
  unsigned char *f = foreground;
  unsigned char fg, cross_fg;
  unsigned char *p = image->pixelPtr;
  unsigned char *q, *r;
  float lat, width, fw;

  coverage_color_copy();

  /*
   * Do we need to draw a blank surround outside
   * the sinusoidal map?
   */
  if ((projection == SINUSOIDAL) && !fgOnly) {
      surround = TRUE;
  } else {
      surround = FALSE;
  }

  fw = (float) w;

  for (i = 0; i < h; i++) {
    if (surround) {
      lat = 90.0 - 180.0 * i / h;
      width = (cos(DEG_TO_RAD * lat) * fw);
      drawWidth = (int) (0.5 + (fw - width) / 2);
    }

    offset = i * w;
    q = p;

    for (j = 0; j < w; j++, p += PIXELSIZE) {
      fg = *(f++);

      if (surround &&
	  (j < drawWidth || j > (w - drawWidth))) {
	p[0] = outline_colors[0];
	p[1] = outline_colors[1];
	p[2] = outline_colors[2];
	q = p; /* updating for cross wrap-around */
	continue;
      }

      if (fg) {
	c = outline_colors + 3 * fg;

	p[0] = c[0];
	p[1] = c[1];
	p[2] = c[2];

        /*
	 * Indicate sub-satellite point by drawing four arms of cross.
	 */
	if ((fg == GROUND_TRACKS_CI) || (fg == SPECIAL_GROUND_TRACKS_CI)) {
	  cross_fg = fg + 1;

	  if (j > 0) {
	    /* draw left arm of cross */
	    *(f-2) = cross_fg; /* set retroactively for future updates. */
	    r = p - PIXELSIZE;
	    r[0] = c[0];
	    r[1] = c[1];
	    r[2] = c[2];
	  }
	  if (i > 0) {
	    /* draw top arm of cross */
	    *(f-w-1) = cross_fg; /* set retroactively for future updates. */
	    r = p - w * PIXELSIZE;
	    r[0] = c[0];
	    r[1] = c[1];
	    r[2] = c[2];
	  }
	  if (j < w-1) {
	    /* ready to draw right arm of cross, next j iteration */
	    *f = cross_fg;
	  } else {
	    /* right arm wrap around - set pixel at j=0 i same value */
	    *(f-j) = cross_fg;
	    q[0] = c[0];
	    q[1] = c[1];
	    q[2] = c[2];
	  }
	  if (i < h-1) {
	    /* ready to draw bottom of cross, next i iteration/row pass */
	    *(f-1 + w) = cross_fg;
	  }
	}

	continue;
      }

      if (!fgOnly) {
	if (!no_access_flag || (g->data[offset + j] > 0)) {
	  ci = MIN(g->data[offset + j], num_colors - 1);
	  c = colors + 3 * ci;
	} else {
	  ci = MIN(g->noaccess[offset + j], num_noaccess_colors - 1);
	  c = noaccess_colors + 3 * ci;
	}
	p[0] = c[0];
	p[1] = c[1];
	p[2] = c[2];
      }
    }
  }

  if (geomview_dynamic_texture && cyl_foreground && texture_flag &&
      (projection == CYLINDRICAL)) {
    write_image_to_scratchfile();
  }
}

/*
 * tracks_to_foreground - place dot on the foreground map which corresponds
 * to the subsatellite point (nadir).  Do this for every satellite
 * in the list.
 */
void
tracks_to_foreground(const Satellite_list SL, int projection, grid * g,
		     unsigned int color_index, const CentralBody * pcb)
{
  unsigned int color;
  Satellite_list sl = SL;

  while (sl) {
    if (sl->s->tag == 1) {
      /* inelegant choice of special color */
      color = SPECIAL_GROUND_TRACKS_CI;
    } else {
      color = color_index;
    }
    track_to_foreground(sl->s, projection, g, color, pcb);
    sl = sl->next;
  }
}

/*
 * track_to_foreground - routine to increment the foreground map at the point
 * underneath the satellite.  For a single satellite.
 *
 * grid_index = color index to insert into the foreground array
 */
void
track_to_foreground(Satellite s, int projection, grid * g,
		    unsigned int color_index, const CentralBody * pcb)
{
  LatLon l;
  int coords[2];

  spherical_to_lat_lon(&l, &(s->x_S), s->t, pcb);
  latlon_to_grid_index(coords, &l, projection, g);
  foreground[g->width * coords[1] + coords[0]] = color_index;

}

/*
 * Given lat/lon coordinates of point on central body, return corresponding
 * index into grid.
 *        grid_index[0]=column index, grid_index[1]=row index.
 */
static void
latlon_to_grid_index(int grid_index[2], LatLon * pl, int projection, grid * g)
{
  double proj[2];
  LatLon l;

  /*
   * The vertical center-line of the grid corresponds to longitude =
   * Longitude_Center_Line, whereas both 2-d projections place longitude = 0
   * along this line.  So shift the input longitude as required.
   */
  l.lat = pl->lat;
  l.lon = pl->lon - Longitude_Center_Line;

  if (l.lon < -180.0) {
    l.lon += 360.0;
  } else if (l.lon > 180.0) {
    l.lon -= 360.0;
  }

  if (projection == CYLINDRICAL) {
    project_latlon_cylindrical(proj, &l);
  } else if (projection == UNPROJECTED) {
    project_latlon_unprojected(proj, &l);
  } else if (projection == SINUSOIDAL) {
    project_latlon_sinusoidal(proj, &l);
  }

  /*
   * It's assumed that proj[0] ranges from -180 to 180, and
   *                   proj[1] ranges from -90 to 90,  no matter what.
   */
  grid_index[0] = ((int) (0.5 + (proj[0] + 180.0) * (g->width) / 360.0));
  grid_index[1] = ((int) (0.5 + proj[1] * (g->height - 1) / 180.0));
}
